import argparse
from pathlib import Path

import torch

from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import (
    Linear1dTranslationArgs,
    LSTMTranslationArgs,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)


def read_commandline_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--build_dir", required=True, type=Path)
    return parser.parse_args()


class LSTMModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.lstm = torch.nn.LSTM(input_size=1, hidden_size=10)
        self.linear = torch.nn.Linear(in_features=10, out_features=1)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        return self.linear(self.lstm(x)[0])


def main() -> None:
    args = read_commandline_args()

    model = LSTMModel()

    fixed_point_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
    work_library_name = "xil_defaultlib"
    translation_args = dict(
        LSTMTranslatable=LSTMTranslationArgs(
            fixed_point_factory=fixed_point_factory,
            sigmoid_resolution=(-2.5, 2.5, 256),
            tanh_resolution=(-1, 1, 256),
            work_library_name=work_library_name,
        ),
        Linear1dTranslatable=Linear1dTranslationArgs(
            fixed_point_factory=fixed_point_factory,
            work_library_name=work_library_name,
        ),
    )

    translatable_layers = translator.translate_model(
        model=model, build_function_mapping=DEFAULT_BUILD_FUNCTION_MAPPING
    )

    code_repr = translator.generate_code(
        translatable_layers=translatable_layers, translation_args=translation_args
    )

    translator.save_code(code_repr=code_repr, path=args.build_dir)


if __name__ == "__main__":
    main()



import os
from dataclasses import dataclass
from typing import Any, Iterable, Iterator

import torch

from elasticai.creator.resource_utils import PathType
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.translator.abstract.translatable import Translatable
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)


@dataclass
class CodeFile:
    file_name: str
    code: Code


@dataclass
class CodeModule:
    module_name: str
    files: Iterable[CodeFile]


def translate_model(
    model: torch.nn.Module,
    build_function_mapping: BuildFunctionMapping = DEFAULT_BUILD_FUNCTION_MAPPING,
) -> Iterator[Translatable]:
    """
    Translates a given PyTorch-model to an intermediate representation. The intermediate representation is represented
    as an iterator of Translatable objects.

    Parameters:
        model (torch.nn.Module): The PyTorch-model that should be translated.
        build_function_mapping (BuildFunctionMapping):
            Object that maps a given PyTorch-layer to its corresponding build function. If not given the default build
            functions will be used.

    Returns:
        Iterator[Translatable]: Iterator that yields for each layer one Translatable object.
    """
    flat_model = filter(
        lambda x: not isinstance(x, torch.nn.Sequential), model.modules()
    )
    for layer in flat_model:
        build_fn = build_function_mapping.get_from_layer(layer)
        if build_fn is not None:
            yield build_fn(layer)


def generate_code(
    translatable_layers: Iterable[Translatable],
    translation_args: dict[str, Any],
) -> Iterator[CodeModule]:
    """
    Takes an iterable of Translatable objects a dictionary of arguments for each Translatable object type and performs
    the translation from the intermediate representation to the actual VHDL code.

    Parameters:
        translatable_layers (Iterable[Translatable]):
            The intermediate representation as an iterator of Translatable objects.
        translation_args (dict[str, dict[str, Any]]):
            Dictionary with the translation arguments for each kind of Translatable included in the translatable_layers.

    Returns:
        Iterator[CodeModule]:
            Iterator of containers that holds the actual VHDL code. The CodeModule objects holds all CodeFile
            objects for one Translatable object (module) and the name of that module. One CodeFile object
            holds the file name of that code files and the actual code as an iterable of str (lines).
    """
    for module_index, module in enumerate(translatable_layers):
        module_class_name = type(module).__name__

        args = translation_args.get(module_class_name)

        components = module.translate(args)
        files = map(lambda x: CodeFile(file_name=x.file_name, code=x()), components)

        yield CodeModule(module_name=f"{module_index}_{module_class_name}", files=files)


def save_code(code_repr: Iterable[CodeModule], path: PathType) -> None:
    """
    Saves the generated code on the file system.

    Parameters:
        code_repr (Iterable[CodeModule]): The generated code that should be saved.
        path (PathType):
            The path to a folder in which the code should be saved. All parent folders that don't exist will be created.
    """
    for module in code_repr:
        module_path = os.path.join(path, module.module_name)
        os.makedirs(module_path, exist_ok=True)

        for code_file in module.files:
            file_path = os.path.join(module_path, code_file.file_name)
            code = "\n".join(code_file.code)

            with open(file_path, "w") as out_file:
                out_file.write(code)



from dataclasses import dataclass, field
from typing import Callable

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.components.utils import (
    calculate_addr_width,
    derive_fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint


@dataclass
class LSTMComponent:
    input_size: int
    hidden_size: int
    fixed_point_factory: Callable[[float], FixedPoint]
    work_library_name: str = field(default="work")

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = derive_fixed_point_params_from_factory(
            self.fixed_point_factory
        )
        self.x_h_addr_width = calculate_addr_width(self.input_size + self.hidden_size)
        self.hidden_addr_width = calculate_addr_width(self.input_size)
        self.w_addr_width = calculate_addr_width(
            (self.input_size + self.hidden_size) * self.hidden_size
        )

    @property
    def file_name(self) -> str:
        return "lstm.vhd"

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "lstm.tpl.vhd")

        code = template.format(
            work_library_name=self.work_library_name,
            data_width=self.data_width,
            frac_width=self.frac_width,
            input_size=self.input_size,
            hidden_size=self.hidden_size,
            x_h_addr_width=self.x_h_addr_width,
            hidden_addr_width=self.hidden_addr_width,
            w_addr_width=self.w_addr_width,
        )

        yield from code.splitlines()



from unittest import TestCase

from elasticai.creator.vhdl.components.dual_port_2_clock_ram_component import (
    DualPort2ClockRamComponent,
)
from elasticai.creator.vhdl.components.lstm_common_component import LSTMCommonComponent
from elasticai.creator.vhdl.components.lstm_component import (
    LSTMComponent as LSTMComponent,
)
from elasticai.creator.vhdl.components.rom_component import RomComponent
from elasticai.creator.vhdl.components.sigmoid_component import SigmoidComponent
from elasticai.creator.vhdl.components.tanh_component import TanhComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers.lstm_translatable import (
    LSTMTranslatable,
    LSTMTranslationArgs,
)


class LSTMTranslatableTest(TestCase):
    def setUp(self) -> None:
        self.lstm = LSTMTranslatable(
            weights_ih=[[[1, 2], [3, 4], [5, 6], [7, 8]]],
            weights_hh=[[[1], [2], [3], [4]]],
            biases_ih=[[1, 2, 3, 4]],
            biases_hh=[[5, 6, 7, 8]],
        )

        self.translation_args = LSTMTranslationArgs(
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=2),
            sigmoid_resolution=(-2.5, 2.5, 256),
            tanh_resolution=(-1, 1, 256),
        )

    def test_correct_number_of_components(self) -> None:
        vhdl_components = list(self.lstm.translate(self.translation_args))
        self.assertEqual(len(vhdl_components), 13)

    def test_contains_all_needed_components(self) -> None:
        vhdl_components = self.lstm.translate(self.translation_args)

        target_components = [
            (RomComponent, "wi_rom.vhd"),
            (RomComponent, "wf_rom.vhd"),
            (RomComponent, "wg_rom.vhd"),
            (RomComponent, "wo_rom.vhd"),
            (RomComponent, "bi_rom.vhd"),
            (RomComponent, "bf_rom.vhd"),
            (RomComponent, "bg_rom.vhd"),
            (RomComponent, "bo_rom.vhd"),
            (SigmoidComponent, "sigmoid.vhd"),
            (TanhComponent, "tanh.vhd"),
            (LSTMComponent, "lstm.vhd"),
            (LSTMCommonComponent, "lstm_common.vhd"),
            (DualPort2ClockRamComponent, "dual_port_2_clock_ram.vhd"),
        ]
        actual_components = [(type(x), x.file_name) for x in vhdl_components]

        self.assertEqual(actual_components, target_components)



from dataclasses import dataclass

from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.precomputed_scalar_function import Tanh


@dataclass
class TanhComponent:
    x: list[FixedPoint]

    @property
    def file_name(self) -> str:
        return f"tanh.vhd"

    def __call__(self) -> Code:
        yield from Tanh(x=self.x, component_name="tanh")()



from elasticai.creator.vhdl.translator.abstract.layers.linear_1d_translatable import (
    Linear1dTranslatable,
    Linear1dTranslationArgs,
)
from elasticai.creator.vhdl.translator.abstract.layers.lstm_translatable import (
    LSTMTranslatable,
    LSTMTranslationArgs,
)



from typing import Iterable, Protocol

from elasticai.creator.resource_utils import Package, read_text
from elasticai.creator.vhdl.language import Code


class VHDLComponent(Protocol):
    @property
    def file_name(self) -> str:
        return ""

    def __call__(self) -> Code:
        ...


class VHDLStaticComponent:
    def __init__(self, template_package: Package, file_name: str) -> None:
        self._template_package = template_package
        self._file_name = file_name

    @property
    def file_name(self) -> str:
        return self._file_name

    def __call__(self) -> Code:
        code = read_text(self._template_package, self._file_name)
        yield from code.splitlines()


VHDLModule = Iterable[VHDLComponent]



import unittest

import torch

from elasticai.creator.vhdl.translator.pytorch.build_functions import build_linear_1d


def arange_parameter(
    start: int, end: int, shape: tuple[int, ...]
) -> torch.nn.Parameter:
    return torch.nn.Parameter(
        torch.reshape(torch.arange(start, end, dtype=torch.float32), shape)
    )


class Linear1dBuildFunctionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.linear = torch.nn.Linear(in_features=3, out_features=1)
        self.linear.weight = arange_parameter(start=1, end=4, shape=(1, -1))
        self.linear.bias = arange_parameter(start=1, end=2, shape=(-1,))

    def test_weights_and_bias_correct_set(self) -> None:
        linear1d = build_linear_1d(self.linear)
        self.assertEqual(linear1d.weight, [[1.0, 2.0, 3.0]])
        self.assertEqual(linear1d.bias, [1.0])



from dataclasses import dataclass, field
from typing import Callable

import numpy as np

from elasticai.creator.vhdl.components import (
    DualPort2ClockRamComponent,
    LSTMCommonComponent,
    LSTMComponent,
    RomComponent,
    SigmoidComponent,
    TanhComponent,
)
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.translatable import Translatable
from elasticai.creator.vhdl.vhdl_component import VHDLModule


@dataclass
class LSTMTranslationArgs:
    fixed_point_factory: Callable[[float], FixedPoint]
    sigmoid_resolution: tuple[float, float, int]
    tanh_resolution: tuple[float, float, int]
    work_library_name: str = field(default="work")


@dataclass
class LSTMTranslatable(Translatable):
    """
    Abstract representation of an LSTM layer that can be directly translated to an iterable of VHDLComponent objects.
    Currently, no stacked LSTMs are supported (only single layer LSTMs are supported).

    Parameters:
        weights_ih (list[list[list[float]]]):
            List of input-hidden weights for each layer. Weights of one layer is a list[list[float]] with the shape
            (4*hidden_size, input_size) and the structure (W_ii | W_if | W_ig | W_io).
        weights_hh (list[list[list[float]]]):
            List of hidden-hidden weights for each layer. Weights of one layer is a list[list[float]] with the shape
            (4*hidden_size, hidden_size) and the structure (W_hi | W_hf | W_hg | W_ho).
        biases_ih (list[list[float]]):
            List of input-hidden biases for each layer. Biases of one layer is a list[float] with the shape
            (4*hidden_size,) and the structure (b_ii | b_if | b_ig | b_io).
        biases_hh (list[list[float]]):
            List of hidden-hidden biases for each layer. Biases of one layer is a list[float] with the shape
            (4*hidden_size,) and the structure (b_hi | b_hf | b_hg | b_ho).
    """

    weights_ih: list[list[list[float]]]
    weights_hh: list[list[list[float]]]
    biases_ih: list[list[float]]
    biases_hh: list[list[float]]

    def _build_weights(
        self, to_fixed_point: Callable[[list[float]], list[FixedPoint]]
    ) -> tuple[tuple[list[FixedPoint], ...], ...]:
        def reshape_params(params: list[list[float]] | np.ndarray) -> np.ndarray:
            return np.array([np.reshape(param, (4, -1)) for param in params])

        weights = np.concatenate((self.weights_ih, self.weights_hh), axis=2)
        weights = reshape_params(weights)
        weights = weights[0]  # Currently only supporting one layer LSTMs
        w_i, w_f, w_g, w_o = weights.reshape(4, -1).tolist()

        bias = np.add(self.biases_ih, self.biases_hh)
        bias = reshape_params(bias)
        bias = bias[0]  # Currently only supporting one layer LSTMs
        b_i, b_f, b_g, b_o = bias.tolist()

        final_weights = tuple(map(to_fixed_point, (w_i, w_f, w_g, w_o)))
        final_biases = tuple(map(to_fixed_point, (b_i, b_f, b_g, b_o)))

        return final_weights, final_biases

    def _derive_input_and_hidden_size(self) -> tuple[int, int]:
        _, hidden_size, input_size = np.shape(self.weights_ih)
        return input_size, hidden_size // 4

    def translate(self, args: LSTMTranslationArgs) -> VHDLModule:
        def to_fp(values: list[float]) -> list[FixedPoint]:
            return list(map(args.fixed_point_factory, values))

        weights, bias = self._build_weights(to_fixed_point=to_fp)
        rom_names = (
            f"{name}_rom" for name in ("wi", "wf", "wg", "wo", "bi", "bf", "bg", "bo")
        )
        for rom_values, rom_name in zip(weights + bias, rom_names):
            yield RomComponent(
                rom_name=rom_name, values=rom_values, resource_option="auto"
            )

        precomputed_sigmoid_inputs = to_fp(np.linspace(*args.sigmoid_resolution).tolist())  # type: ignore
        precomputed_tanh_inputs = to_fp(np.linspace(*args.tanh_resolution).tolist())  # type: ignore
        yield SigmoidComponent(x=precomputed_sigmoid_inputs)
        yield TanhComponent(x=precomputed_tanh_inputs)

        input_size, hidden_size = self._derive_input_and_hidden_size()
        yield LSTMComponent(
            input_size=input_size,
            hidden_size=hidden_size,
            fixed_point_factory=args.fixed_point_factory,
            work_library_name=args.work_library_name,
        )

        yield LSTMCommonComponent()
        yield DualPort2ClockRamComponent()



import torch

from elasticai.creator.vhdl.translator.abstract.layers import Linear1dTranslatable


def build_linear_1d(linear: torch.nn.Linear) -> Linear1dTranslatable:
    def to_list(tensor: torch.Tensor) -> list:
        return tensor.detach().numpy().tolist()

    return Linear1dTranslatable(
        weight=to_list(linear.weight), bias=to_list(linear.bias)
    )



from elasticai.creator.vhdl.translator.pytorch.build_functions.linear_1d_build_function import (
    build_linear_1d,
)
from elasticai.creator.vhdl.translator.pytorch.build_functions.lstm_build_function import (
    build_lstm,
)



import unittest
from dataclasses import dataclass
from typing import Any, Iterable

import torch.nn

from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.translator.abstract.layers import LSTMTranslatable
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)
from elasticai.creator.vhdl.translator.pytorch.translator import CodeFile, CodeModule
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


@dataclass
class VHDLComponentMock:
    name: str
    code: list[str]

    @property
    def file_name(self) -> str:
        return self.name

    def __call__(self) -> Code:
        yield from self.code


@dataclass
class TranslatableMock:
    components: list[VHDLComponent]

    def translate(self, args: Any) -> VHDLModule:
        yield from self.components


def fake_build_function(module: torch.nn.Module) -> TranslatableMock:
    return TranslatableMock(
        components=[
            VHDLComponentMock(name="component1", code=["1", "2", "3"]),
            VHDLComponentMock(name="component2", code=["4", "5", "6"]),
        ]
    )


def unpack_module_directories(
    modules: Iterable[CodeModule],
) -> list[tuple[str, list[tuple[str, Code]]]]:
    def unpack_code_file(code_file: CodeFile) -> tuple[str, Code]:
        return code_file.file_name, list(code_file.code)

    return [
        (module.module_name, list(map(unpack_code_file, module.files)))
        for module in modules
    ]


class TranslatorTest(unittest.TestCase):
    def setUp(self) -> None:
        self.build_mapping = BuildFunctionMapping(
            mapping={"torch.nn.modules.rnn.LSTM": fake_build_function}
        )

    def test_translate_model_empty_model(self) -> None:
        model = torch.nn.Sequential()
        translated_model = translator.translate_model(model, self.build_mapping)
        self.assertEqual(len(list(translated_model)), 0)

    def test_translate_model_with_one_layer(self) -> None:
        model = torch.nn.Sequential(torch.nn.LSTM(input_size=1, hidden_size=2))

        translated_model = list(translator.translate_model(model, self.build_mapping))

        self.assertEqual(len(translated_model), 1)
        self.assertEqual(type(translated_model[0]), TranslatableMock)

    def test_generate_code(self) -> None:
        model = torch.nn.Sequential(torch.nn.LSTM(input_size=1, hidden_size=2))
        translated_model = translator.translate_model(model, self.build_mapping)
        modules = translator.generate_code(
            translatable_layers=translated_model, translation_args=dict()
        )

        code = unpack_module_directories(modules)
        expected_code = [
            (
                "0_TranslatableMock",
                [
                    ("component1", ["1", "2", "3"]),
                    ("component2", ["4", "5", "6"]),
                ],
            )
        ]

        self.assertEqual(code, expected_code)

    @unittest.SkipTest
    def test_translate_model_correct_ordering_of_layers(self):
        class Model(torch.nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.lstm_2 = torch.nn.LSTM(input_size=2, hidden_size=3)
                self.lstm_1 = torch.nn.LSTM(input_size=1, hidden_size=2)

            def forward(self, x: torch.Tensor) -> torch.Tensor:
                return self.lstm_2(self.lstm_1(x))

        def extract_input_hidden_size(lstm: LSTMTranslatable) -> tuple[int, int]:
            hidden_size = len(lstm.weights_hh[0][0])
            input_size = len(lstm.weights_ih[0][0])
            return input_size, hidden_size

        model = Model()
        translated = list(
            translator.translate_model(model, DEFAULT_BUILD_FUNCTION_MAPPING)
        )

        self.assertEqual(extract_input_hidden_size(translated[0]), (1, 2))  # type: ignore
        self.assertEqual(extract_input_hidden_size(translated[1]), (2, 3))  # type: ignore



import unittest

from elasticai.creator.vhdl.components import (
    Linear1dComponent,
    LSTMCommonComponent,
    RomComponent,
)
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import (
    Linear1dTranslatable,
    Linear1dTranslationArgs,
)


class Linear1dTranslatableTest(unittest.TestCase):
    def setUp(self) -> None:
        self.linear = Linear1dTranslatable(weight=[[1, 2, 3]], bias=[1])
        self.translation_args = Linear1dTranslationArgs(
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4)
        )

    def test_contains_all_needed_components(self) -> None:
        vhdl_components = self.linear.translate(self.translation_args)

        target_components = [
            (Linear1dComponent, "linear_1d.vhd"),
            (RomComponent, "w_rom.vhd"),
            (RomComponent, "b_rom.vhd"),
            (LSTMCommonComponent, "lstm_common.vhd"),
        ]
        actual_components = [(type(x), x.file_name) for x in vhdl_components]

        self.assertEqual(actual_components, target_components)



from elasticai.creator.vhdl.vhdl_component import VHDLStaticComponent


class LSTMCommonComponent(VHDLStaticComponent):
    def __init__(self):
        super().__init__(
            template_package="elasticai.creator.vhdl.templates",
            file_name="lstm_common.vhd",
        )
